; Programa: Package para entorno para agente de juego a Bumper Cars
;           ver (http://www.discover.com/nov_01/bogglers.html)
;
; Asignatura: Introduccin a la Inteligencia Artificial
;             ETSII - ULL
; Prctica: 6a (sistemas de produccin)
; Profesor: Patricio Garca (patricio@csi.ull.es)
; Fecha: 13/5/2003
; Comentarios:
;           Para representar el laberinto se utilizan estructuras
;           tipo bump (ver (help 'bump)).
;           Las funciones accesibles mas importantes son:
;	       inic-bump :   Comienza a ejecutar el agente
;              gen-bump :    Devuelve estructura con estado de inicio
;              vis-bump :    Visualiza estructura tipo bump
;              version-bump: Devuelve el nmero de versin
;
;           Ver el correspondiente help para cada una de ellas.
;
;           Para poder utilizarse este paquete ha de cargarse mediante:
;              (load "bumper")
;              (use-package "bumper")
;
; ToDo list:
;           - Probar bajo otros lisps
;           - Chekeos de consistencia de datos de entrada
;

(make-package "bumper" :nicknames '(bmp) :use '("LISP"))
(in-package "bumper")

(export '(version-bumper *sol-ordenada-bump* *sol-simultanea-bump* *obs-bump* *max-cars-bump* bump make-bump bump-mat bump-cars bump-exits bump-p gen-bump  vis-bump  mov-val-bump mov-bump sol-val-bump inic-bump))


; Variables globales
; ------------------

(defvar *version-bumper* "0.4" "Versin de bumper.lsp")

(defun version-bumper ()
  " - Funcin: VERSION-BUMPER ()
     Package:BUMPER

Imprime la versin del paquete bumper que se ha cargado"
  *version-bumper*)

(defvar *sol-ordenada-bump* nil "Indica tipo de solucin que se busca: t, cada coche en su salida; nil, cada coche en una salida cualquiera")

(defvar *sol-simultanea-bump* nil "Indica tipo de problema: t, hay que llegar simultaneamente con los coches a las salidas; nil, no es necesario, una vez un coche llega a una salida se detiene") 

(defvar verbose-bump nil "Activador de visualizacin traza de depuracin")


; Estructura de representacin del problema a resolver
; ----------------------------------------------------

(defvar *obs-bump* 'obs
  " - Variable: TIPO smbolo 

Contiene el smbolo que representa los obstculos (#) dentro de la matiz del campo MAT de la estructura tipo BUMP.")

(defvar *max-cars-bump* 5 
  "- Variable: TIPO nmero 

Contiene el nmero mximo de coches y salidas que puede tener una estructura tipo BUMP.")

(defstruct bump
  " - Estructura: BUMP (DEFSTRUCT MAT CARS EXITS)
     Package:BUMPER

Estructura que almacena un laberinto Bumper, junto a las posiciones de los coches y las salidas (5 de cada mximo). Consta de tres campos: MAT es una matriz de dos dimensiones, y CARS y EXITS que son listas.

La matriz MAT representa el laberinto y tiene por lo tanto las dimensiones mximas de ste. Las paredes del laberinto (#) se representa con el smbolo contenido en la variable *OBS-BUMP* (ver (help '*OBS-BUMP*))'. Los espacios vacios estan representados mediante NIL. Las salidas se representan con nmeros negativos (-1, -2, ...).

El campo CARS almacena una lista con tantas sublistas como coches existen, indicando la fila y columna donde esta cada coche.

El campo EXITS guarda la informacin de las salidas, al igua que CARS consta de una lista con sublistas de tantos elemento como salidas indicando fila y columna de la salida. Ntese que este campo es redundantes, ya que tambin est representados en la matriz MAT.

La mayora de funciones de este paquete hace uso de este tipo de estructura."
  mat
  cars
  exits)


; Funciones para creacion de la estructura BUMP
; ---------------------------------------------

(defun trcar (car)
  (cond
    ((eq car #\#) *obs-bump*)
    ((eq car #\Space) nil)
    ((digit-char-p car) (digit-char-p car))
    (t (- (char-code #\a) (char-code car) 1))))

(defun line2list (line)
  (if (stringp line)
      (let (list)
	(dotimes (i (length line) (reverse list))
	  (setf list (cons (trcar (aref line i)) list))))
    line))

(defun elim-plusp (elm)
  (if (or (not (numberp elm))
	  (not (plusp elm)))
      elm))

(defun cribalist (list)
  (mapcar #'elim-plusp list))

(defun listnnil (n)
  (if (zerop n)
      nil
    (cons nil (listnnil (1- n)))))

(defun completa (cols list)
  (dolist (sublis list list)
    (nconc sublis (listnnil (- cols (length sublis))))))

(defun crea-est (listinic cars-def exits-def maxfil maxcol)
  "Crea e iniciaciza estructura bump"
  (make-bump :cars cars-def 
	     :exits exits-def
	     :mat (make-array 
		   (list maxfil maxcol)
		   :initial-contents (completa maxcol listinic))))


; Carga problema por defecto o desde fichero
; ------------------------------------------

(defun vec-to-list (vec n)
  (let (list)
  (dotimes (i n (reverse list))
    (push (aref vec i) list))))

(defun load-lab-bump (infile)
  (do* ((list (line2list (read-line infile nil 'eof)) 
	      (line2list (read-line infile nil 'eof)))
	(fil 0 (1+ fil))
	(maxcol 0)
	(listmat nil)
	(num-cars 0)
	(vec-cars (make-array (list *max-cars-bump*)))
	(vec-exits (make-array (list *max-cars-bump*))))
      ((eq list 'eof) (crea-est (reverse listmat) 
				(vec-to-list vec-cars num-cars)
				(vec-to-list vec-exits num-cars)
				fil maxcol))
    ; Busca coches y salidas
    (dolist (num (remove-if-not #'numberp list) nil)
      (setf num-cars (max num-cars (abs num)))
      (if (plusp num)
          ; Es coche
	  (setf (aref vec-cars (1- num)) (list fil (position num list)))
	; Es salida
	(setf (aref vec-exits (- -1 num)) (list fil (position num list)))))
    ; Actualiza mximo de columnas y guarda fila
    (setf maxcol (max maxcol (length list)))
    (setf listmat (cons (cribalist list) listmat))))

(defun load-lab-fich-bump (fichero)
  ; Carga desde fichero
  (with-open-file (infile fichero)
		  (load-lab-bump infile)))

(defun load-lab-str-bump (string)
  ; Carga desde string
  (let ((infile (make-string-input-stream string)))
    (load-lab-bump infile)))


; Generacion estructura inicial
; -----------------------------

(defvar str-def 
" #####
##a#b## 
#     #
#  # ##
# #   #
#   # #
# 1 2 #
#######"
  "String con laberinto defecto")

(defun gen-bump-def ()
  "Genera laberinto por defecto y/o internos"
  (load-lab-str-bump str-def))

(defun num-fich (numero)
  (format nil "~Alab~D.bump" pathaddhof numero))

(defvar fichero nil
  "Fichero de laberinto a cargar")

(defun gen-bump ()
  " - Funcin: GEN-BUMP ()
     Package:BUMPER

Genera un estado inicial con un laberinto para el Bumper Cars, de acuerdo a las especificiaciones del argumento FICH pasado a la funcin INIC-BUMP, puede cargar el laberinto desde un fichero si asi se le ha especificado en esta variable (ver (help 'INIC-BUMP). Devuelve una estructura tipo BUMP."
  (cond
   ((null fichero) (gen-bump-def))
    ((numberp fichero) (load-lab-fich-bump (num-fich fichero)))
    (t (load-lab-fich-bump fichero))))


; Visualizacin de estructura de problema
; ---------------------------------------

(defun num-car (num)
  (code-char (+ (char-code #\a) num)))

(defun car-o-exit (pos-car pos-exit)
  (cond
   (pos-car (1+ pos-car))
   (pos-exit (num-car pos-exit))
   (t " ")))

(defun vis-bump (est)
  " - Funcin: VIS-BUMP (EST)
     Package:BUMPER

Visualiza el estado EST utlizando caracteres: 1 y 2 para los coches, a y b para las salidas, # para los obstculos (paredes). EST ha de ser una estructura tipo BUMP."
  (dotimes (i (array-dimension (bump-mat est) 0) t)
    (terpri)
    (dotimes (j (array-dimension (bump-mat est) 1) t)
      (if (eq (aref (bump-mat est) i j) *obs-bump*)
	  (princ #\#)
	(princ (car-o-exit (position `(,i ,j) (bump-cars est) :test #'equal)
			   (position `(,i ,j) (bump-exits est) :test #'equal)))))))


; Cargamos Chequeador de soluciones
; ---------------------------------
(load "check-bump")  ; Disponible en fichero check-bump.o


; Lanzador de agentes y problemas
; -------------------------------

(defun getlistpartidas (fich)
  "Obtiene lista de particas a partir de fich"
  (cond
   ((stringp fich) (directory fich))
   ((numberp fich) (list fich))
   ((null fich) '(()))
   ((listp fich) (apply #'nconc (mapcar #'getlistpartidas fich)))
   (t '(()))))


(defun gen-tipo (fich ord sim)
  "Genera el tipo de problema a insertar en hof"
  (format nil "~A_~D"
	  fich
	  (1+ (position `(,ord ,sim)
			'((NIL NIL) (NIL T) (T T) (T NIL)) 
			:test #'equal))))


; Comienza a ejecutar el agente pasado
; ------------------------------------
; Parmetros: 
;
; Variables globales afectadas:
;             verbose-bump
;             fichero
;             *sol-ordenada-bump*
;             *son-simultanea-bump*
;
; Valor de retorno:
;             (XITOS LONG-SOLS TIEMPO-EJEC )  
;                 Todo funcion correctamente
;             NIL Hubo algn problema
;
; Requerimientos:
;           * Exista una funcion AGENTE que encuentre una solucin
;             vlida al problema y devuelva los pasos de la solucin
;           * Si se instancia FICH, que sea un string empareje el 
;             nombre de fichero(s) existente(s) que contenga(n) 
;             laberinto(s)
;
; Ejemplo llamada: (INIC-BUMP :tipo 2) -> (1 13 12)

(defun inic-bump (&optional
			 (agente #'agente-humano) ;jugador
			 ; debug        ; modo debug
			 &key 
			 verb         ; modo verbose
			 (num nil)    ; Numero de partidas a jugar
			 (fich nil)   ; Fichero/s de partidas
			 (ord nil)    ; Solucin ordenada
			 (sim nil)    ; solucin simultnea
			 hof)         ; Aadir resultado al H.O.F.
  
  " - Funcin: INIC-BUMP (&OPTIONAL (AGENTE #'agente-humano) &KEY VERB (NUM 1) (FICH NIL) (ORD NIL) (SIM NIL) HOF)
     Package:BUMPER

Inicia la ejecucion del agente AGENTE que juega al Bumper. La funcion AGENTE debe encontrar una solucion y devolverla como una lista de movimientos, cada movimiento puede ser 1, 2, 3  4 segn sean norte, sur, este u oeste. VERB indica si se activa el modo verbose. NUM indica el nmero de partidas a jugar. FICH indica el laberinto de juego, puede ser NIL (laberinto por defecto), o bien un string que cargar el laberinto desde el/los fichero(s) que empareje. Dichos ficheros son de texto, donde los obstculos viene representados por el caracter '#', los espacios vacios por espacios en blanco, los coches por dgitos a partir del '1' y los destnos por letras desde la 'a' en adelante. ORD indica si se desea que en la solucin cada coche quede ubicado en su destino correspondiente (T) o no (NIL). SIM indica si se buscan soluciones en que los coches llegen simultaneamente a los destinos, y por tanto los coches no han de pararse al llegar a un destino (t), o bien se soluciones no simultaneas, en las que los coches al llegar a su destino se quedan congelados en l (NIL).HOF indica si se graban los resultados en el Hall-of-Fame (/scratch/bumper/hof.txt , ver por*.txt).
Devuelve una lista con el nmero de partidas solucionadas con xito, el nmero acumulado de pasos de las soluciones encontradas, y el tiempo de ejecucion total que han tardado las partidas."

  (setf verbose-bump verb)
  (setf fichero fich)
  (setf *sol-ordenada-bump* ord)
  (setf *sol-simultanea-bump* sim)
  (let (tiempo-ejec
	(listpart (getlistpartidas fich))
	listparact
	sol
	long-sol
	(long-sols 0)
	(fallos 0)
	(exitos 0))
    (if (null num)
	(setf num (length listpart)))
    (setf tiempo-ejec (get-internal-run-time))
    (dotimes (partida num)
      (if (null listparact)
	  (setf listparact listpart))
      (setf fichero (pop listparact))
      (setf sol (funcall agente))  ;Llamada al agente
      (setf long-sol (check-bump sol))
      (if (null long-sol)
	  (progn
	    (setf fallos (1+ fallos))
	    (format t "~%El agente fallo!!")))
      (if verbose-bump
	  (progn
	    (format t "~%Partida:          ~D~%" (1+ partida))
	    (format t "Long. solucion:   ~D~%" long-sol)
	    (format t "Solucion:         ~s~%" sol)))
      (if long-sol
	  (setf long-sols (+ long-sols long-sol))))
    (setf tiempo-ejec (- (get-internal-run-time) tiempo-ejec))
    (setf exitos (- num fallos))
    (if verbose-bump
	(progn
	  (format t "~%Nmero de partidas:            ~D~%" num)
	  (format t "Nmero de xitos:              ~D~%" exitos)
	  (format t "Tipo de partidas:              ~D~%" fich)
	  (format t "Tipo de problemas:             ~D~%" (list ord sim))
	  (if (plusp exitos)
	      (progn
		(format t "Promedio Long. sol.:           ~D~%" (float (/ long-sols exitos)))
		(format t "Tiempo promedio por partida:   ~D~%~%" (float (/ tiempo-ejec exitos)))))))
    (if (and hof (plusp exitos))
      (printhof (gen-tipo fich *sol-ordenada-bump* *sol-simultanea-bump*)
		agente 
		long-sols
		tiempo-ejec 
		exitos))
    (list exitos long-sols tiempo-ejec)))


